@@ -31,6 +31,7 @@ using System;
 using System.Data;
 using System.Text;
 using System.Collections.Generic;
+using System.Linq;
 
 using Mono.Unix;
 
@@ -42,42 +43,43 @@ using Hyena.Query;
 using Banshee.Base;
 using Banshee.Query;
 using Banshee.Database;
+using Banshee.PlaybackController;
 
 namespace Banshee.Collection.Database
-{       
-    public class DatabaseTrackListModel : TrackListModel, IExportableModel, 
+{
+    public class DatabaseTrackListModel : TrackListModel, IExportableModel,
         ICacheableDatabaseModel, IFilterable, ISortable, ICareAboutView
     {
         private readonly BansheeDbConnection connection;
         private IDatabaseTrackModelProvider provider;
         protected IDatabaseTrackModelCache cache;
         private Banshee.Sources.DatabaseSource source;
-        
+
         private long count;
 
         private long filtered_count;
         private TimeSpan filtered_duration;
-        private long filtered_filesize;
-        
+        private long filtered_filesize, filesize;
+
         private ISortableColumn sort_column;
         private string sort_query;
         private bool forced_sort_query;
-        
+
         private string reload_fragment;
-        private string join_table, join_fragment, join_primary_key, join_column, condition;
+        private string join_table, join_fragment, join_primary_key, join_column, condition, condition_from;
 
         private string query_fragment;
         private string user_query;
 
         private int rows_in_view;
-        
+
         public DatabaseTrackListModel (BansheeDbConnection connection, IDatabaseTrackModelProvider provider, Banshee.Sources.DatabaseSource source)
         {
             this.connection = connection;
             this.provider = provider;
             this.source = source;
         }
-        
+
         protected HyenaSqliteConnection Connection {
             get { return connection; }
         }
@@ -93,7 +95,7 @@ namespace Banshee.Collection.Database
             cache.AggregatesUpdated += HandleCacheAggregatesUpdated;
             GenerateSortQueryPart ();
         }
-        
+
         private bool have_new_user_query = true;
         private void GenerateUserQueryFragment ()
         {
@@ -130,36 +132,31 @@ namespace Banshee.Collection.Database
         {
             SortQuery = (SortColumn == null || SortColumn.SortType == SortType.None)
                 ? (SortColumn != null && source is Banshee.Playlist.PlaylistSource)
-                    ? "CorePlaylistEntries.ViewOrder ASC, CorePlaylistEntries.EntryID ASC" 
+                    ? "CorePlaylistEntries.ViewOrder ASC, CorePlaylistEntries.EntryID ASC"
                     : BansheeQuery.GetSort ("Artist", true)
                 : BansheeQuery.GetSort (SortColumn.SortKey, SortColumn.SortType == SortType.Ascending);
         }
 
-        public void Sort (ISortableColumn column)
+        private SortType last_sort_type = SortType.None;
+        public bool Sort (ISortableColumn column)
         {
             lock (this) {
                 if (forced_sort_query) {
-                    return;
+                    return false;
                 }
-                
-                if (sort_column == column && sort_column != null) {
-                    switch (sort_column.SortType) {
-                        case SortType.Ascending:    sort_column.SortType = SortType.Descending; break;
-                        case SortType.Descending:   sort_column.SortType = SortType.None; break;
-                        case SortType.None:         sort_column.SortType = SortType.Ascending; break;
-                    }
+
+                // Don't sort by the same column and the same sort-type more than once
+                if (sort_column != null && sort_column == column && column.SortType == last_sort_type) {
+                    return false;
                 }
-                
-                // If we're switching from a different column, or we aren't in a playlist, make sure sort type isn't None
-                if (sort_column != null && (sort_column != column || !(source is Banshee.Playlist.PlaylistSource)) && sort_column.SortType == SortType.None) {
-                    sort_column.SortType = SortType.Ascending;
-                }
-                
+
+                last_sort_type = column.SortType;
                 sort_column = column;
-            
+
                 GenerateSortQueryPart ();
                 cache.Clear ();
             }
+            return true;
         }
 
         private void HandleCacheAggregatesUpdated (IDataReader reader)
@@ -167,43 +164,61 @@ namespace Banshee.Collection.Database
             filtered_duration = TimeSpan.FromMilliseconds (reader.IsDBNull (1) ? 0 : Convert.ToInt64 (reader[1]));
             filtered_filesize = reader.IsDBNull (2) ? 0 : Convert.ToInt64 (reader[2]);
         }
-        
+
         public override void Clear ()
         {
             cache.Clear ();
             count = 0;
+            filesize = 0;
             filtered_count = 0;
             OnCleared ();
         }
 
-        public void InvalidateCache ()
+        public void InvalidateCache (bool notify)
         {
             if (cache == null) {
                 Log.ErrorFormat ("Called invalidate cache for {0}'s track model, but cache is null", source);
             } else {
                 cache.Clear ();
-                OnReloaded ();
+                if (notify) {
+                    OnReloaded ();
+                }
             }
         }
 
         private string unfiltered_query;
-        protected string UnfilteredQuery {
+        public virtual string UnfilteredQuery {
             get {
-                return unfiltered_query ?? unfiltered_query = String.Format (
-                    "FROM {0}{1} WHERE {2} {3}",
-                    provider.From, JoinFragment,
+                return unfiltered_query ?? (unfiltered_query = String.Format (
+                    "FROM {0} WHERE {1} {2}",
+                    FromFragment,
                     String.IsNullOrEmpty (provider.Where) ? "1=1" : provider.Where,
                     ConditionFragment
-                );
+                ));
             }
         }
 
-        protected virtual void UpdateUnfilteredAggregates ()
+        private string from;
+        protected string From {
+            get { return from ?? provider.From; }
+            set { from = value; }
+        }
+
+        private string from_fragment;
+        public string FromFragment {
+            get { return from_fragment ?? (from_fragment = String.Format ("{0}{1}", From, JoinFragment)); }
+        }
+
+        public virtual void UpdateUnfilteredAggregates ()
         {
             HyenaSqliteCommand count_command = new HyenaSqliteCommand (String.Format (
-                "SELECT COUNT(*) {0}", UnfilteredQuery
+                "SELECT COUNT(*), SUM(CoreTracks.FileSize) {0}", UnfilteredQuery
             ));
-            count = connection.Query<long> (count_command);
+
+            using (HyenaDataReader reader = new HyenaDataReader (connection.Query (count_command))) {
+                count = reader.Get<long> (0);
+                filesize = reader.Get<long> (1);
+            }
         }
 
         public override void Reload ()
@@ -243,12 +258,12 @@ namespace Banshee.Collection.Database
                     foreach (IFilterListModel model in reload_models) {
                         model.Reload (false);
                     }
-                    
+
                     bool have_filters = false;
                     foreach (IFilterListModel filter in source.CurrentFilters) {
                         have_filters |= !filter.Selection.AllSelected;
                     }
-                    
+
                     // Unless both artist/album selections are "all" (eg unfiltered), reload
                     // the track model again with the artist/album filters now in place.
                     if (have_filters) {
@@ -279,7 +294,7 @@ namespace Banshee.Collection.Database
         {
             StringBuilder qb = new StringBuilder ();
             qb.Append (UnfilteredQuery);
-            
+
             if (with_filters) {
                 foreach (IFilterListModel filter in source.CurrentFilters) {
                     string filter_sql = filter.GetSqlFilter ();
@@ -289,47 +304,50 @@ namespace Banshee.Collection.Database
                     }
                 }
             }
-            
+
             if (query_fragment != null) {
                 qb.Append (" AND ");
                 qb.Append (query_fragment);
             }
-            
+
             if (sort_query != null) {
                 qb.Append (" ORDER BY ");
                 qb.Append (sort_query);
             }
-            
+
             reload_fragment = qb.ToString ();
             cache.Reload ();
         }
 
         public override int IndexOf (TrackInfo track)
         {
-            return (int) cache.IndexOf (track as DatabaseTrackInfo);
+            lock (this) {
+                if (track is DatabaseTrackInfo) {
+                    return (int) cache.IndexOf (track as DatabaseTrackInfo);
+                } else if (track is Banshee.Streaming.RadioTrackInfo) {
+                    return (int) cache.IndexOf ((track as Banshee.Streaming.RadioTrackInfo).ParentTrack as DatabaseTrackInfo);
+                }
+                return -1;
+            }
         }
 
-        private DateTime random_began_at = DateTime.MinValue;
-        private DateTime last_random = DateTime.MinValue;
-        private static string random_fragment = "AND (LastPlayedStamp < ? OR LastPlayedStamp IS NULL) AND (LastSkippedStamp < ? OR LastSkippedStamp IS NULL) ORDER BY RANDOM()";
-        public override TrackInfo GetRandom (DateTime notPlayedSince, bool repeat)
+        public int IndexOfFirst (TrackInfo track)
         {
             lock (this) {
-                if (Count == 0)
-                    return null;
+                return IndexOf (cache.GetSingle ("AND MetadataHash = ? ORDER BY OrderID", track.MetadataHash));
+            }
+        }
 
-                if (random_began_at < notPlayedSince)
-                    random_began_at = last_random = notPlayedSince;
+        public override TrackInfo GetRandom (DateTime notPlayedSince)
+        {
+            return GetRandom (notPlayedSince, PlaybackShuffleMode.Song, true, false, Shuffler.Playback);
+        }
 
-                TrackInfo track = cache.GetSingle (random_fragment, random_began_at, random_began_at);
-
-                if (track == null && repeat) {
-                    random_began_at = last_random;
-                    track = cache.GetSingle (random_fragment, random_began_at, random_began_at);
-                }
-
-                last_random = DateTime.Now;
-                return track;
+        public TrackInfo GetRandom (DateTime notPlayedSince, PlaybackShuffleMode mode, bool repeat, bool resetSinceTime, Shuffler shuffler)
+        {
+            lock (this) {
+                shuffler.SetModelAndCache (this, cache);
+                return shuffler.GetRandom (notPlayedSince, mode, repeat, resetSinceTime);
             }
         }
 
@@ -352,7 +370,11 @@ namespace Banshee.Collection.Database
         public long FileSize {
             get { return filtered_filesize; }
         }
-        
+
+        public long UnfilteredFileSize {
+            get { return filesize; }
+        }
+
         public int UnfilteredCount {
             get { return (int) count; }
             set { count = value; }
@@ -360,20 +382,22 @@ namespace Banshee.Collection.Database
 
         public string UserQuery {
             get { return user_query; }
-            set { 
+            set {
                 lock (this) {
-                    user_query = value; 
+                    user_query = value;
                     have_new_user_query = true;
                 }
             }
         }
-        
+
         public string ForcedSortQuery {
             get { return forced_sort_query ? sort_query : null; }
-            set { 
+            set {
                 forced_sort_query = value != null;
                 sort_query = value;
-                cache.Clear ();
+                if (cache != null) {
+                    cache.Clear ();
+                }
             }
         }
 
@@ -398,16 +422,46 @@ namespace Banshee.Collection.Database
             get { return join_column; }
             set { join_column = value; }
         }
-        
+
         public void AddCondition (string part)
         {
-            condition = condition == null ? part : String.Format ("{0} AND {1}", condition, part);
+            AddCondition (null, part);
+        }
+
+        public void AddCondition (string tables, string part)
+        {
+            if (!String.IsNullOrEmpty (part)) {
+                condition = condition == null ? part : String.Format ("{0} AND {1}", condition, part);
+
+                if (!String.IsNullOrEmpty (tables)) {
+                    condition_from = condition_from == null ? tables : String.Format ("{0}, {1}", condition_from, tables);
+                }
+            }
         }
 
         public string Condition {
             get { return condition; }
         }
 
+        private string condition_from_fragment;
+        public string ConditionFromFragment {
+            get {
+                if (condition_from_fragment == null) {
+                    if (JoinFragment == null) {
+                        condition_from_fragment = condition_from;
+                    } else {
+                        if (condition_from == null) {
+                            condition_from = "CoreTracks";
+                        }
+
+                        condition_from_fragment = String.Format ("{0}{1}", condition_from, JoinFragment);
+                    }
+                }
+
+                return condition_from_fragment;
+            }
+        }
+
         public string ConditionFragment {
             get { return PrefixCondition ("AND"); }
         }
@@ -424,20 +478,20 @@ namespace Banshee.Collection.Database
             get { return (int) cache.CacheId; }
         }
 
-        public ISortableColumn SortColumn { 
+        public ISortableColumn SortColumn {
             get { return sort_column; }
         }
-                
+
         public virtual int RowsInView {
             protected get { return rows_in_view; }
             set { rows_in_view = value; }
         }
 
-        int IExportableModel.GetLength () 
+        int IExportableModel.GetLength ()
         {
             return Count;
         }
-        
+
         IDictionary<string, object> IExportableModel.GetMetadata (int index)
         {
             return this[index].GenerateExportable ();
@@ -479,7 +533,7 @@ namespace Banshee.Collection.Database
         public string ReloadFragment {
             get { return reload_fragment; }
         }
-        
+
         public bool CachesValues { get { return false; } }
     }
 }
